==> dovecot-0.92.tar.gz: dovecot-0.92/src/lib-storage/subscribe-file/subscribe-file.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ugly code here - text files are annoying to manage */

#include "lib.h"
#include "mmap-util.h"
#include "imap-util.h"
#include "mail-storage.h"
#include "subscribe-file.h"

#include <unistd.h>
#include <fcntl.h>

#define SUBSCRIPTION_FILE_NAME ".subscriptions"

static int lock_file(int fd, int type)
{
	struct flock fl;

	/* lock whole file */
	fl.l_type = type;
	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 0;

	while (fcntl(fd, F_SETLKW, &fl) == -1) {
		if (errno != EINTR)
			return FALSE;
	}

	return TRUE;
}

static int subscription_open(MailStorage *storage, int update,
			     const char **path, void **mmap_base,
			     size_t *mmap_length)
{
	int fd;

	*path = t_strconcat(storage->dir, "/" SUBSCRIPTION_FILE_NAME, NULL);

	fd = update ? open(*path, O_RDWR | O_CREAT, 0660) :
		open(*path, O_RDONLY | O_CREAT, 0660);
	if (fd == -1) {
		mail_storage_set_critical(storage, "Can't open subscription "
					  "file %s: %s", *path, strerror(errno));
		return -1;
	}

	if (!lock_file(fd, update ? F_WRLCK : F_RDLCK)) {
		mail_storage_set_critical(storage, "fcntl() failed for "
					  "subscription file %s: %s", *path,
					  strerror(errno));
		(void)close(fd);
		return -1;
	}

	*mmap_base = update ? mmap_rw_file(fd, mmap_length) :
		mmap_ro_file(fd, mmap_length);
	if (*mmap_base == MAP_FAILED) {
		*mmap_base = NULL;
		mail_storage_set_critical(storage, "mmap() failed for "
					  "subscription file %s: %s", *path,
					  strerror(errno));
		(void)close(fd);
		return -1;
	}

	(void)madvise(*mmap_base, *mmap_length, MADV_SEQUENTIAL);
	return fd;
}

static int subscription_append(MailStorage *storage, int fd, const char *name,
			       int len, int append_lf, const char *path)
{
	char *buf;

	if (lseek(fd, 0, SEEK_END) == (off_t)-1) {
		mail_storage_set_critical(storage, "lseek() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		return FALSE;
	}

	buf = t_buffer_get(len+2);
	buf[0] = '\n';
	memcpy(buf+1, name, len);
	buf[len+1] = '\n';

	if (append_lf)
		len += 2;
	else {
		buf++;
		len++;
	}

	if (write(fd, buf, len) != len) {
		mail_storage_set_critical(storage, "write() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		return FALSE;
	}

	return TRUE;
}

int subsfile_set_subscribed(MailStorage *storage, const char *name, int set)
{
	void *mmap_base;
	size_t mmap_length;
	const char *path;
	char *subscriptions, *end, *p;
	int fd, namelen, afterlen, removelen, failed, append_lf;

	if (strcasecmp(name, "INBOX") == 0)
		name = "INBOX";

	fd = subscription_open(storage, TRUE, &path, &mmap_base, &mmap_length);
	if (fd == -1)
		return FALSE;

	namelen = strlen(name);

	subscriptions = mmap_base;
	if (subscriptions == NULL)
		p = NULL;
	else {
		end = subscriptions + mmap_length;
		for (p = subscriptions; p != end; p++) {
			if (*p == *name && p+namelen <= end &&
			    strncmp(p, name, namelen) == 0) {
				/* make sure beginning and end matches too */
				if ((p == subscriptions || p[-1] == '\n') &&
				    (p+namelen == end || p[namelen] == '\n'))
					break;
			}
		}

		if (p == end)
			p = NULL;
	}

	failed = FALSE;
	if (p != NULL && !set) {
		/* remove it */
		afterlen = mmap_length - (int) (p - subscriptions);
		removelen = namelen < afterlen ? namelen+1 : namelen;

		if (removelen < afterlen)
			memmove(p, p+removelen, afterlen-removelen);

		if (ftruncate(fd, mmap_length - removelen) == -1) {
			mail_storage_set_critical(storage,
						  "ftruncate() failed for "
						  "subscription file %s: %s",
						  path, strerror(errno));
			failed = TRUE;
		}
	} else if (p == NULL && set) {
		/* append it */
		append_lf = mmap_length > 0 &&
			subscriptions[mmap_length-1] != '\n';
		if (!subscription_append(storage, fd, name, namelen,
					 append_lf, path))
			failed = TRUE;
	}

	if (mmap_base != NULL && munmap(mmap_base, mmap_length) == -1) {
		mail_storage_set_critical(storage, "munmap() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		failed = TRUE;
	}

	if (close(fd) == -1) {
		mail_storage_set_critical(storage, "close() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		failed = TRUE;
	}
	return !failed;
}

int subsfile_foreach(MailStorage *storage, const char *mask,
		     SubsFileForeachFunc func, void *user_data)
{
	const char *path, *start, *end, *p, *line;
	void *mmap_base;
	size_t mmap_length;
	int fd, ret;

	fd = subscription_open(storage, FALSE, &path, &mmap_base, &mmap_length);
	if (fd == -1)
		return -1;

	start = mmap_base; end = start + mmap_length; ret = 1;
	while (ret) {
		t_push();

		for (p = start; p != end; p++) {
			if (*p == '\n')
				break;
		}

		line = t_strndup(start, (int) (p-start));
		if (line != NULL && *line != '\0' &&
		    imap_mailbox_mask_match(line, mask, storage->hierarchy_sep))
			ret = func(storage, line, user_data);
		t_pop();

		if (p == end)
			break;
		start = p+1;
	}

	(void)close(fd);
	return ret;
}

==> dovecot-0.95.tar.gz: dovecot-0.95/src/lib-storage/subscribe-file/subscribe-file.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ugly code here - text files are annoying to manage */

#include "lib.h"
#include "mmap-util.h"
#include "imap-match.h"
#include "mail-storage.h"
#include "subscribe-file.h"

#include <unistd.h>
#include <fcntl.h>

#define SUBSCRIPTION_FILE_NAME ".subscriptions"

static int lock_file(int fd, int type)
{
	struct flock fl;

	/* lock whole file */
	fl.l_type = type;
	fl.l_whence = SEEK_SET;
	fl.l_start = 0;
	fl.l_len = 0;

	while (fcntl(fd, F_SETLKW, &fl) == -1) {
		if (errno != EINTR)
			return FALSE;
	}

	return TRUE;
}

static int subscription_open(MailStorage *storage, int update,
			     const char **path, void **mmap_base,
			     size_t *mmap_length)
{
	int fd;

	*path = t_strconcat(storage->dir, "/" SUBSCRIPTION_FILE_NAME, NULL);

	fd = update ? open(*path, O_RDWR | O_CREAT, 0660) :
		open(*path, O_RDONLY | O_CREAT, 0660);
	if (fd == -1) {
		mail_storage_set_critical(storage, "Can't open subscription "
					  "file %s: %s", *path, strerror(errno));
		return -1;
	}

	if (!lock_file(fd, update ? F_WRLCK : F_RDLCK)) {
		mail_storage_set_critical(storage, "fcntl() failed for "
					  "subscription file %s: %s", *path,
					  strerror(errno));
		(void)close(fd);
		return -1;
	}

	*mmap_base = update ? mmap_rw_file(fd, mmap_length) :
		mmap_ro_file(fd, mmap_length);
	if (*mmap_base == MAP_FAILED) {
		*mmap_base = NULL;
		mail_storage_set_critical(storage, "mmap() failed for "
					  "subscription file %s: %s", *path,
					  strerror(errno));
		(void)close(fd);
		return -1;
	}

	(void)madvise(*mmap_base, *mmap_length, MADV_SEQUENTIAL);
	return fd;
}

static int subscription_append(MailStorage *storage, int fd, const char *name,
			       int len, int append_lf, const char *path)
{
	char *buf;

	if (lseek(fd, 0, SEEK_END) == (off_t)-1) {
		mail_storage_set_critical(storage, "lseek() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		return FALSE;
	}

	buf = t_buffer_get(len+2);
	buf[0] = '\n';
	memcpy(buf+1, name, len);
	buf[len+1] = '\n';

	if (append_lf)
		len += 2;
	else {
		buf++;
		len++;
	}

	if (write(fd, buf, len) != len) {
		mail_storage_set_critical(storage, "write() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		return FALSE;
	}

	return TRUE;
}

int subsfile_set_subscribed(MailStorage *storage, const char *name, int set)
{
	void *mmap_base;
	size_t mmap_length;
	const char *path;
	char *subscriptions, *end, *p;
	int fd, namelen, afterlen, removelen, failed, append_lf;

	if (strcasecmp(name, "INBOX") == 0)
		name = "INBOX";

	fd = subscription_open(storage, TRUE, &path, &mmap_base, &mmap_length);
	if (fd == -1)
		return FALSE;

	namelen = strlen(name);

	subscriptions = mmap_base;
	if (subscriptions == NULL)
		p = NULL;
	else {
		end = subscriptions + mmap_length;
		for (p = subscriptions; p != end; p++) {
			if (*p == *name && p+namelen <= end &&
			    strncmp(p, name, namelen) == 0) {
				/* make sure beginning and end matches too */
				if ((p == subscriptions || p[-1] == '\n') &&
				    (p+namelen == end || p[namelen] == '\n'))
					break;
			}
		}

		if (p == end)
			p = NULL;
	}

	failed = FALSE;
	if (p != NULL && !set) {
		/* remove it */
		afterlen = mmap_length - (int) (p - subscriptions);
		removelen = namelen < afterlen ? namelen+1 : namelen;

		if (removelen < afterlen)
			memmove(p, p+removelen, afterlen-removelen);

		if (ftruncate(fd, mmap_length - removelen) == -1) {
			mail_storage_set_critical(storage,
						  "ftruncate() failed for "
						  "subscription file %s: %s",
						  path, strerror(errno));
			failed = TRUE;
		}
	} else if (p == NULL && set) {
		/* append it */
		append_lf = mmap_length > 0 &&
			subscriptions[mmap_length-1] != '\n';
		if (!subscription_append(storage, fd, name, namelen,
					 append_lf, path))
			failed = TRUE;
	}

	if (mmap_base != NULL && munmap(mmap_base, mmap_length) == -1) {
		mail_storage_set_critical(storage, "munmap() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		failed = TRUE;
	}

	if (close(fd) == -1) {
		mail_storage_set_critical(storage, "close() failed for "
					  "subscription file %s: %s", path,
					  strerror(errno));
		failed = TRUE;
	}
	return !failed;
}

int subsfile_foreach(MailStorage *storage, const char *mask,
		     SubsFileForeachFunc func, void *user_data)
{
        const ImapMatchGlob *glob;
	const char *path, *start, *end, *p, *line;
	void *mmap_base;
	size_t mmap_length;
	int fd, ret;

	fd = subscription_open(storage, FALSE, &path, &mmap_base, &mmap_length);
	if (fd == -1)
		return -1;

	glob = imap_match_init(mask, TRUE, storage->hierarchy_sep);

	start = mmap_base; end = start + mmap_length; ret = 1;
	while (ret) {
		t_push();

		for (p = start; p != end; p++) {
			if (*p == '\n')
				break;
		}

		line = t_strndup(start, (int) (p-start));
		if (line != NULL && *line != '\0' &&
		    imap_match(glob, line, 0, NULL) >= 0)
			ret = func(storage, line, user_data);
		t_pop();

		if (p == end)
			break;
		start = p+1;
	}

	(void)close(fd);
	return ret;
}
